@@ -361,6 +361,7 @@ xt = _LazyModule.load('xonsh', 'xonsh.tools', 'xt')
 xp = _LazyModule.load('xonsh', 'xonsh.platform', 'xp')
 # amalgamated xonsh.completers.tools
 SKIP_TOKENS = {'sudo', 'time', 'timeit', 'which', 'showcmd', 'man'}
+END_PROC_TOKENS = {'|', '||', '&&', 'and', 'or'}
 
 
 def complete_command(cmd, line, start, end, ctx):
@@ -387,27 +388,32 @@ def complete_skipper(cmd, line, start, end, ctx):
     Skip over several tokens (e.g., sudo) and complete based on the rest of the
     line.
     """
-    res = line.split(' ', 1)
-    if len(res) == 2:
-        first, rest = res
-    else:
-        first = res[0]
-        rest = ''
-    if first in SKIP_TOKENS:
-        comp = builtins.__xonsh_shell__.shell.completer
-        res = rest.split(' ', 1)
-        if len(res) == 1:
-            comp_func = complete_command
-        else:
-            comp_func = comp.complete
-        return comp_func(cmd,
-                         rest,
-                         start - len(first) - 1,
-                         end - len(first) - 1,
-                         ctx)
-    else:
+    parts = line.split(' ')
+    skip_part_num = 0
+    for i, s in enumerate(parts):
+        if s in END_PROC_TOKENS:
+            skip_part_num = i + 1
+    while len(parts) > skip_part_num:
+        if parts[skip_part_num] not in SKIP_TOKENS:
+            break
+        skip_part_num += 1
+
+    if skip_part_num == 0:
         return set()
 
+    if len(parts) == skip_part_num + 1:
+        comp_func = complete_command
+    else:
+        comp = builtins.__xonsh_shell__.shell.completer
+        comp_func = comp.complete
+
+    skip_len = len(' '.join(line[:skip_part_num])) + 1
+    return comp_func(cmd,
+                     ' '.join(parts[skip_part_num:]),
+                     start - skip_len,
+                     end - skip_len,
+                     ctx)
+
 #
 # man
 #
@@ -472,6 +478,7 @@ def complete_from_man(prefix, line, start, end, ctx):
 # path
 #
 # amalgamated os
+# amalgamated re
 ast = _LazyModule.load('ast', 'ast')
 # amalgamated builtins
 # amalgamated xonsh.tools
@@ -479,11 +486,12 @@ ast = _LazyModule.load('ast', 'ast')
 # amalgamated xonsh.lazyasd
 # amalgamated xonsh.completers.tools
 @xl.lazyobject
-def CHARACTERS_NEED_QUOTES():
-    cnq = ' `\t\r\n${}*()"\',?&'
+def PATTERN_NEED_QUOTES():
+    pattern = r'\s`\$\{\}\,\*\(\)"\'\?&'
     if xp.ON_WINDOWS:
-        cnq += '%'
-    return cnq
+        pattern += '%'
+    pattern = '[' + pattern + ']' + r'|\band\b|\bor\b'
+    return re.compile(pattern)
 
 
 def _path_from_partial_string(inp, pos=None):
@@ -605,7 +613,7 @@ def _quote_paths(paths, start, end):
         start = orig_start
         end = orig_end
         if (start == '' and
-                (any(i in s for i in CHARACTERS_NEED_QUOTES) or
+                (re.search(PATTERN_NEED_QUOTES, s) is not None or
                  (backslash in s and slash != backslash))):
             start = end = _quote_to_use(s)
         if os.path.isdir(expand_path(s)):
@@ -657,7 +665,8 @@ def _splitpath_helper(path, sofar=()):
     folder, path = os.path.split(path)
     if path:
         sofar = sofar + (path, )
-    if not folder or folder == xt.get_sep():
+    if (not folder or folder == xt.get_sep() or
+       (xp.ON_WINDOWS and os.path.splitdrive(path)[0])):
         return sofar[::-1]
     return _splitpath_helper(folder, sofar)
 
@@ -796,6 +805,21 @@ def XONSH_TOKENS():
 
 
 def complete_python(prefix, line, start, end, ctx):
+    """
+    Completes based on the contents of the current Python environment,
+    the Python built-ins, and xonsh operators.
+    If there are no matches, split on common delimiters and try again.
+    """
+    rtn = _complete_python(prefix, line, start, end, ctx)
+    if not rtn:
+        prefix = re.split(r'\(|=|{|\[|,', prefix)[-1]
+        start = line.find(prefix)
+        rtn = _complete_python(prefix, line, start, end, ctx)
+        return rtn, len(prefix)
+    return rtn
+
+
+def _complete_python(prefix, line, start, end, ctx):
     """
     Completes based on the contents of the current Python environment,
     the Python built-ins, and xonsh operators.
@@ -822,6 +846,8 @@ def complete_python_mode(prefix, line, start, end, ctx):
         return set()
     prefix_start = prefix[:2]
     python_matches = complete_python(prefix[2:], line, start+2, end, ctx)
+    if isinstance(python_matches, cabc.Sequence):
+        python_matches = python_matches[0]
     return set(prefix_start + i for i in python_matches)
 
 
@@ -907,6 +933,7 @@ def complete_module(prefix):
 # base
 #
 """Base completer for xonsh."""
+# amalgamated collections.abc
 # amalgamated xonsh.completers.path
 # amalgamated xonsh.completers.python
 # amalgamated xonsh.completers.commands
@@ -921,13 +948,20 @@ def complete_base(prefix, line, start, end, ctx):
         paths = complete_path(prefix, line, start, end, ctx, False)
         return (out | paths[0]), paths[1]
     elif prefix == line:
-        return (complete_python(prefix, line, start, end, ctx) |
-                complete_command(prefix, line, start, end, ctx))
+        python_comps = complete_python(prefix, line, start, end, ctx)
+        if isinstance(python_comps, cabc.Sequence):
+            return (python_comps[0] |
+                    complete_command(prefix, line, start, end, ctx),
+                    python_comps[1])
+        else:
+            return (python_comps |
+                    complete_command(prefix, line, start, end, ctx))
     return set()
 
 #
 # bash
 #
+# amalgamated os
 shlex = _LazyModule.load('shlex', 'shlex')
 pathlib = _LazyModule.load('pathlib', 'pathlib')
 # amalgamated builtins
@@ -936,17 +970,29 @@ pathlib = _LazyModule.load('pathlib', 'pathlib')
 # amalgamated xonsh.completers.path
 BASH_COMPLETE_SCRIPT = r"""
 {sources}
-if (complete -p "{cmd}" 2> /dev/null || echo _minimal) | grep --quiet -e "_minimal"
+
+function _get_complete_statement {{
+    complete -p {cmd} 2> /dev/null || echo "-F _minimal"
+}}
+
+_complete_stmt=$(_get_complete_statement)
+if echo "$_complete_stmt" | grep --quiet -e "_minimal"
 then
-    declare -f _completion_loader > /dev/null && _completion_loader "{cmd}"
+    declare -f _completion_loader > /dev/null && _completion_loader {cmd}
+    _complete_stmt=$(_get_complete_statement)
 fi
-_func=$(complete -p {cmd} | grep -o -e '-F \w\+' | cut -d ' ' -f 2)
+
+_func=$(echo "$_complete_stmt" | grep -o -e '-F \w\+' | cut -d ' ' -f 2)
+declare -f "$_func" > /dev/null || exit 1
+
+echo "$_complete_stmt"
 COMP_WORDS=({line})
 COMP_LINE={comp_line}
 COMP_POINT=${{#COMP_LINE}}
 COMP_COUNT={end}
 COMP_CWORD={n}
 $_func {cmd} {prefix} {prev}
+
 for ((i=0;i<${{#COMPREPLY[*]}};i++)) do echo ${{COMPREPLY[i]}}; done
 """
 
@@ -970,27 +1016,45 @@ def complete_from_bash(prefix, line, begidx, endidx, ctx):
             if idx >= begidx:
                 break
         prev = tok
+
     if len(prefix) == 0:
-        prefix = '""'
+        prefix_quoted = '""'
         n += 1
     else:
-        prefix = shlex.quote(prefix)
+        prefix_quoted = shlex.quote(prefix)
 
     script = BASH_COMPLETE_SCRIPT.format(
         sources='\n'.join(sources), line=' '.join(shlex.quote(p) for p in splt),
-        comp_line=shlex.quote(line), n=n, cmd=cmd,
-        end=endidx + 1, prefix=prefix, prev=shlex.quote(prev),
+        comp_line=shlex.quote(line), n=n, cmd=shlex.quote(cmd),
+        end=endidx + 1, prefix=prefix_quoted, prev=shlex.quote(prev),
     )
 
     try:
         out = subprocess.check_output(
-            [xp.bash_command()], input=script, universal_newlines=True,
+            [xp.bash_command(), '-c', script], universal_newlines=True,
             stderr=subprocess.PIPE, env=builtins.__xonsh_env__.detype())
     except (subprocess.CalledProcessError, FileNotFoundError):
-        out = ''
+        return set()
 
-    rtn = _quote_paths(set(out.splitlines()), '', '')
-    return rtn
+    out = out.splitlines()
+    complete_stmt = out[0]
+    out = set(out[1:])
+
+    # From GNU Bash document: The results of the expansion are prefix-matched
+    # against the word being completed
+    commprefix = os.path.commonprefix(out)
+    strip_len = 0
+    while strip_len < len(prefix):
+        if commprefix.startswith(prefix[strip_len:]):
+            break
+        strip_len += 1
+
+    if '-o noquote' not in complete_stmt:
+        out = _quote_paths(out, '', '')
+    if '-o nospace' in complete_stmt:
+        out = set([x.rstrip() for x in out])
+
+    return out, len(prefix) - strip_len
 
 
 def _collect_completions_sources():
