@@ -1,4 +1,5 @@
 import os
+import re
 import ast
 import builtins
 
@@ -10,11 +11,12 @@ from xonsh.completers.tools import get_filter_function
 
 
 @xl.lazyobject
-def CHARACTERS_NEED_QUOTES():
-    cnq = ' `\t\r\n${}*()"\',?&'
+def PATTERN_NEED_QUOTES():
+    pattern = r'\s`\$\{\}\,\*\(\)"\'\?&'
     if xp.ON_WINDOWS:
-        cnq += '%'
-    return cnq
+        pattern += '%'
+    pattern = '[' + pattern + ']' + r'|\band\b|\bor\b'
+    return re.compile(pattern)
 
 
 def _path_from_partial_string(inp, pos=None):
@@ -136,7 +138,7 @@ def _quote_paths(paths, start, end):
         start = orig_start
         end = orig_end
         if (start == '' and
-                (any(i in s for i in CHARACTERS_NEED_QUOTES) or
+                (re.search(PATTERN_NEED_QUOTES, s) is not None or
                  (backslash in s and slash != backslash))):
             start = end = _quote_to_use(s)
         if os.path.isdir(expand_path(s)):
@@ -188,7 +190,8 @@ def _splitpath_helper(path, sofar=()):
     folder, path = os.path.split(path)
     if path:
         sofar = sofar + (path, )
-    if not folder or folder == xt.get_sep():
+    if (not folder or folder == xt.get_sep() or
+       (xp.ON_WINDOWS and os.path.splitdrive(path)[0])):
         return sofar[::-1]
     return _splitpath_helper(folder, sofar)
 
