@@ -128,6 +128,11 @@ def ctrl_d_condition(cli):
                 not cli.current_buffer.text)
 
 
+@Condition
+def autopair_condition(cli):
+    return builtins.__xonsh_env__.get("XONSH_AUTOPAIR", False)
+
+
 def can_compile(src):
     """Returns whether the code can be compiled, i.e. it is valid xonsh."""
     src = src if src.endswith('\n') else src + '\n'
@@ -169,6 +174,53 @@ def load_xonsh_bindings(key_bindings_manager):
         """ Insert literal tab on Shift+Tab instead of autocompleting """
         event.cli.current_buffer.insert_text(env.get('INDENT'))
 
+    @handle('(', filter=autopair_condition)
+    def insert_right_parens(event):
+        event.cli.current_buffer.insert_text('(')
+        event.cli.current_buffer.insert_text(')', move_cursor=False)
+
+    @handle(')', filter=autopair_condition)
+    def overwrite_right_parens(event):
+        buffer = event.cli.current_buffer
+        if buffer.document.current_char == ')':
+            buffer.cursor_position += 1
+        else:
+            buffer.insert_text(')')
+
+    @handle('[', filter=autopair_condition)
+    def insert_right_bracket(event):
+        event.cli.current_buffer.insert_text('[')
+        event.cli.current_buffer.insert_text(']', move_cursor=False)
+
+    @handle(']', filter=autopair_condition)
+    def overwrite_right_bracket(event):
+        buffer = event.cli.current_buffer
+
+        if buffer.document.current_char == ']':
+            buffer.cursor_position += 1
+        else:
+            buffer.insert_text(']')
+
+    @handle('\'', filter=autopair_condition)
+    def insert_right_quote(event):
+        buffer = event.cli.current_buffer
+
+        if buffer.document.current_char == '\'':
+            buffer.cursor_position += 1
+        else:
+            buffer.insert_text('\'')
+            buffer.insert_text('\'', move_cursor=False)
+
+    @handle('"', filter=autopair_condition)
+    def insert_right_double_quote(event):
+        buffer = event.cli.current_buffer
+
+        if buffer.document.current_char == '"':
+            buffer.cursor_position += 1
+        else:
+            buffer.insert_text('"')
+            buffer.insert_text('"', move_cursor=False)
+
     @handle(Keys.ControlD, filter=ctrl_d_condition)
     def call_exit_alias(event):
         """Use xonsh exit function"""
